# -*- coding: utf-8 -*-

from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from requests import Request as HTTPRequest, Session
from requests.cookies import create_cookie, RequestsCookieJar
from urllib.parse import parse_qs, urlparse, urlencode
from requests_toolbelt.utils import dump

from model_utils.models import TimeStampedModel

from smithy.helpers import render_with_context, parse_dump_result


class NameValueModel(TimeStampedModel):
    name = models.CharField(max_length = 200)
    value = models.TextField()

    def __str__(self):
        return self.name

    class Meta:
        abstract = True


class Request(TimeStampedModel):
    """
    A base model shared by RequestBlueprint and
    RequestRecord. Used solely to reduce
    """
    METHODS = (
        ('GET', 'GET'),
        ('POST', 'POST'),
        ('PUT', 'PUT'),
        ('DELETE', 'DELETE'),
        ('OPTIONS', 'OPTIONS'),
        ('HEAD', 'HEAD'),
    )
    BODY_TYPES = (
        ('', 'Other'),
        ('application/x-www-form-urlencoded', 'x-www-form-urlencoded'),
        ('application/json', 'JSON'),
        ('text/plain', 'Text'),
        ('application/javascript', 'JavaScript'),
        ('application/xml', 'XML (application/xml)'),
        ('text/xml', 'XML (text/xml)'),
        ('text/html', 'HTML'),
    )
    method = models.CharField(
        max_length = 7, choices = METHODS,
        blank = False, null = False)
    name = models.CharField(max_length = 500, blank = False)
    url = models.CharField(max_length = 2083)
    body = models.TextField()
    content_type = models.CharField(
        default = BODY_TYPES[0][0],
        blank = True, null = True,
        max_length = 100, choices = BODY_TYPES)

    def __str__(self):
        if self.name:
            return self.name
        return "{} {}".format(
            self.method,
            self.url,
        )


class RequestBlueprint(Request):
    """
    A blueprint for HTTP requests. This model will be
    used to generate and send HTTP requests. Once sent,
    a RequestRecord will be created for that request.
    """
    follow_redirects = models.BooleanField(
        default = False, blank = False, null = False)

    @property
    def name_value_related(self):
        return [
            self.headers,
            self.query_parameters,
            self.cookies
        ]

    def send(self, context = None):

        context = context or {}
        for variable in self.variables.all():
            context[variable.name] = variable.value

        request = HTTPRequest(
            url = render_with_context(self.url, context),
            method = self.method)

        session = Session()
        record = RequestRecord.objects.create(blueprint = self)

        # Copy RequestBlueprint values to RequestRecord
        for qs in self.name_value_related:
            for obj in qs.all():
                obj.pk = 0
                obj.name = render_with_context(obj.name, context)
                obj.value = render_with_context(obj.value, context)
                obj.add_to(request)
                obj.request = record
                obj.save()

        if self.content_type == self.BODY_TYPES[0][0]:
            data = render_with_context(self.body, context)
        else:
            data = {}
            for param in self.body_parameters.all():
                param.add_to(data, context)

        request.data = data
        prepared_request = request.prepare()

        response = session.send(prepared_request, stream = True)
        # TODO: follow redirects

        RequestRecord.objects.filter(pk = record.pk).update(
            raw_request = parse_dump_result(dump._dump_request_data, prepared_request),
            raw_response = parse_dump_result(dump._dump_response_data, response),
            status = response.status_code,
            **RequestRecord.get_clone_args(self, context)
        )

        # Return fresh copy after update
        return RequestRecord.objects.get(pk = record.pk)


class RequestRecord(Request):
    """
    A record of a Request that has been sent.
    Contains response and diagnostic information
    about the request.
    """
    raw_request = models.TextField()
    raw_response = models.TextField()
    status = models.PositiveIntegerField(null = True)
    blueprint = models.ForeignKey(
        'smithy.RequestBlueprint',
        on_delete = models.SET_NULL,
        null = True)

    @property
    def is_success(self):
        return self.status and self.status < 400

    @staticmethod
    def of_blueprint(blueprint):
        return RequestRecord.objects.filter(
            blueprint = blueprint)

    @classmethod
    def get_clone_args(cls, obj, context : dict):
        return dict([
            (
                render_with_context(fld.name, context),
                render_with_context(getattr(obj, fld.name), context)
            )
            for fld \
            in cls._meta.fields \
            if fld.name != obj._meta.pk \
            and fld in obj._meta.fields \
            and fld.name not in [
                   'request', 'id', 'created', 'updated'
               ]])


class Variable(NameValueModel):
    """
    A static variable for this request. Will be added
    directly to the request's context. Use this to
    add custom context to this request that may be
    used in several places.
    """
    request = models.ForeignKey(
        'smithy.RequestBlueprint',
        on_delete = models.CASCADE,
        related_name = 'variables')


class BodyParameter(NameValueModel):
    """
    A x-www-form-urlencoded parameter. Will be properly
    urlencoded and added to the body when the request is
    sent.
    """
    request = models.ForeignKey(
        'smithy.RequestBlueprint',
        on_delete = models.CASCADE,
        related_name = 'body_parameters')

    def add_to(self, payload : dict, context: dict):
        if self.name and self.value:
            payload[
                render_with_context(self.name, context)
            ] = render_with_context(self.value, context)
        else:
            pass # TODO: warn


class Header(NameValueModel):
    """
    An HTTP header. HTTP headers are case insensitive,
    but to ensure consistent behaviour, header names
    will be converted to uppercase before rqeuests
    are sent.
    """
    request = models.ForeignKey(
        'smithy.Request',
        on_delete = models.CASCADE,
        related_name = 'headers')

    def add_to(self, request : HTTPRequest):
        if self.name and self.value:
            request.headers[self.name] = self.value
        else:
            pass # TODO: warn


class QueryParameter(NameValueModel):
    """
    A query parameter. Static query parameters can be
    set in the RequestBlueprint's URL field, but it may
    be easier to manage large or complex sets of query
    parameters separately.
    """
    request = models.ForeignKey(
        'smithy.Request',
        on_delete = models.CASCADE,
        related_name = 'query_parameters')

    def add_to(self, request : HTTPRequest):
        existing_param_str = urlparse(request.url).query
        param_dict = parse_qs(existing_param_str)

        param_dict[self.name] = self.value

        base_url = request.url.rstrip(existing_param_str)
        param_str = urlencode(param_dict, doseq=True)
        request.url = base_url.rstrip("?") + "?" + param_str


class Cookie(NameValueModel):
    """
    A convenience method to avoid needing to define the
    Cookie header and follow correct syntax. If a request
    has both a Cookie defined and a Header with the name
    "Cookie", the Header will used and the Cookie(s) will
    be ignored.
    """
    request = models.ForeignKey(
        'smithy.Request',
        on_delete = models.CASCADE,
        related_name = 'cookies')

    def add_to(self, request : HTTPRequest):
        request.cookies = request.cookies or RequestsCookieJar()
        request.cookies.set_cookie(
            create_cookie(self.name, self.value)
        )


@receiver(post_save, sender = RequestBlueprint)
def add_content_type(sender, instance : RequestBlueprint, created, **kwargs):
    if created and not instance.headers.filter(name__iexact = 'content-type').exists():
        instance.content_type and Header.objects.create(
            name = 'Content-Type',
            value = instance.content_type,
            request = instance.request_ptr
        )



## -*- coding: utf-8 -*-
from typing import List, Union

from django.contrib import admin, messages
from django.db import models
from django.db.models import QuerySet
from django.forms.widgets import TextInput
from django.conf import settings
from smithy.models import RequestBlueprint, RequestRecord, Header, QueryParameter, Cookie, Variable, BodyParameter

CODEMIRROR_PATH = getattr(settings, 'SMITHY_CODEMIRROR_PATH', "https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.44.0/").rstrip('/')


def send(modeladmin, request, queryset : Union[QuerySet, List[RequestBlueprint]]):
    for obj in queryset:
        obj.send()
    messages.success(request, "Sent {} request{}".format(
        queryset.count(),
        's' if queryset.count() != 1 else ''
    ))

send.short_description = "Send"


def ObjectInline(m, readonly = False):
    class Inline(admin.TabularInline):
        extra = 0
        max_num = 0 if readonly else None
        model = m
        can_delete = not readonly
        fields = ['name', 'value']
        formfield_overrides = {
            models.TextField: {'widget': TextInput},
        }
        def get_readonly_fields(self, request, obj = None):
            if obj and readonly:
                self.readonly_fields = self.fields
            return self.readonly_fields

    return Inline


class RequestAdmin(admin.ModelAdmin):
    list_display = ['name', 'url', 'method', 'created', 'modified']
    exclude = ['request_ptr', 'id']
    fields = [
        'name',
        'method',
        #'follow_redirects',
        'url',
        'content_type',
        'body',
    ]
    inlines = [
        ObjectInline(BodyParameter),
        ObjectInline(Header),
        ObjectInline(QueryParameter),
        ObjectInline(Cookie),
        ObjectInline(Variable)
    ]


class RequestBlueprintAdmin(RequestAdmin):
    actions = [
        send,
    ]

    class Media:
        css = {
            'all': (CODEMIRROR_PATH + '/codemirror.css',)
        }

        js = (
            CODEMIRROR_PATH + '/codemirror.js',
            CODEMIRROR_PATH + '/addon/mode/overlay.js',
            CODEMIRROR_PATH + '/mode/xml/xml.js',
            CODEMIRROR_PATH + '/mode/htmlmixed/htmlmixed.js',
            CODEMIRROR_PATH + '/mode/django/django.js',
            'js/smithy.js',
        )


class RequestRecordAdmin(RequestAdmin):
    inlines = [
        ObjectInline(Header, True),
        ObjectInline(QueryParameter, True),
        ObjectInline(Cookie, True),
    ]

    fields = RequestAdmin.fields + ['raw_request', 'raw_response']

    def get_readonly_fields(self, request, obj=None):
        if obj:
            self.readonly_fields = [
                field.name for field in obj.__class__._meta.fields
                if field.name not in self.exclude]
        return self.readonly_fields

    def get_formsets_with_inlines(self, request, obj=None):
        for inline in self.get_inline_instances(request, obj):
            if obj is not None and inline.get_queryset(request).count() > 0:
                yield inline.get_formset(request, obj), inline

    class Media:
        css = {
            'all': ('css/smithy.css',)
        }


admin.site.register(RequestBlueprint, RequestBlueprintAdmin)
admin.site.register(RequestRecord, RequestRecordAdmin)



# Generated by Django 2.1.5 on 2019-03-16 20:32

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Cookie',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=200)),
                ('value', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Header',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=200)),
                ('value', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='QueryParameter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=200)),
                ('value', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Request',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('method', models.CharField(choices=[('GET', 'GET'), ('POST', 'POST'), ('PUT', 'PUT'), ('DELETE', 'DELETE'), ('OPTIONS', 'OPTIONS'), ('HEAD', 'HEAD')], max_length=7)),
                ('name', models.CharField(max_length=500)),
                ('url', models.CharField(max_length=2083)),
                ('body', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Variable',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=200)),
                ('value', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestBlueprint',
            fields=[
                ('request_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='smithy.Request')),
                ('follow_redirects', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
            bases=('smithy.request',),
        ),
        migrations.CreateModel(
            name='RequestRecord',
            fields=[
                ('request_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='smithy.Request')),
                ('response', models.TextField()),
                ('blueprint', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='smithy.RequestBlueprint')),
            ],
            options={
                'abstract': False,
            },
            bases=('smithy.request',),
        ),
        migrations.AddField(
            model_name='queryparameter',
            name='request',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='query_parameters', to='smithy.Request'),
        ),
        migrations.AddField(
            model_name='header',
            name='request',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='headers', to='smithy.Request'),
        ),
        migrations.AddField(
            model_name='cookie',
            name='request',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='cookies', to='smithy.Request'),
        ),
        migrations.AddField(
            model_name='variable',
            name='request',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='variables', to='smithy.RequestBlueprint'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-03-17 17:14

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    replaces = [('smithy', '0002_auto_20190317_1052'), ('smithy', '0003_auto_20190317_1103'), ('smithy', '0004_auto_20190317_1104'), ('smithy', '0005_auto_20190317_1107'), ('smithy', '0006_request_body_type'), ('smithy', '0007_auto_20190317_1159'), ('smithy', '0008_auto_20190317_1213')]

    dependencies = [
        ('smithy', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='BodyParameter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=200)),
                ('value', models.TextField()),
                ('request', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='body_parameters', to='smithy.RequestBlueprint')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='requestrecord',
            name='raw_request',
            field=models.TextField(default=''),
        ),
        migrations.RemoveField(
            model_name='requestrecord',
            name='response',
        ),
        migrations.AddField(
            model_name='requestrecord',
            name='status',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AddField(
            model_name='requestrecord',
            name='raw_response',
            field=models.TextField(default=''),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='request',
            name='content_type',
            field=models.CharField(blank=True, choices=[('', 'Other'), ('application/x-www-form-urlencoded', 'x-www-form-urlencoded'), ('application/json', 'JSON'), ('text/plain', 'Text'), ('application/javascript', 'JavaScript'), ('application/xml', 'XML (application/xml)'), ('text/xml', 'XML (text/xml)'), ('text/html', 'HTML')], default='', max_length=100, null=True),
        ),
    ]



from django.template import Template, Context
from requests_toolbelt.utils import dump

def render_with_context(template, context):
    template = Template(template)
    context = Context(context)
    return template.render(context)

def parse_dump_result(fun, obj):
    prefixes = dump.PrefixSettings('', '')
    try:
        result = bytearray()
        fun(obj, prefixes, result)
        return result.decode('utf-8')
    except Exception:
        return "Could not parse request as a string"
